"""
Basic MS COFF object loader based on https://docs.microsoft.com/en-us/windows/win32/debug/pe-format
"""

from __future__ import annotations

import ctypes
import logging
import struct
from enum import IntEnum, IntFlag

import archinfo

from cle.utils import extract_null_terminated_bytestr

from .backend import Backend, register_backend
from .region import Section, Segment
from .relocation import Relocation
from .symbol import Symbol, SymbolType

log = logging.getLogger(__name__)


class IMAGE_FILE_MACHINE(IntEnum):
    """
    Machine Types
    """

    I386 = 0x14C
    AMD64 = 0x8664


class CoffFileHeader(ctypes.LittleEndianStructure):
    """
    COFF File Header
    """

    _pack_ = 1
    _fields_ = [
        ("Machine", ctypes.c_uint16),
        ("NumberOfSections", ctypes.c_uint16),
        ("TimeDateStamp", ctypes.c_uint32),
        ("PointerToSymbolTable", ctypes.c_uint32),
        ("NumberOfSymbols", ctypes.c_uint32),
        ("SizeOfOptionalHeader", ctypes.c_uint16),
        ("Characteristics", ctypes.c_uint16),
    ]


class IMAGE_SCN(IntFlag):
    """
    Section Flags (Characteristics field)
    """

    MEM_EXECUTE = 0x20000000
    MEM_READ = 0x40000000
    MEM_WRITE = 0x80000000
    CNT_UNINITIALIZED_DATA = 0x00000080


class CoffSectionTableEntry(ctypes.LittleEndianStructure):
    """
    COFF Section Header
    """

    _pack_ = 1
    _fields_ = [
        ("Name", ctypes.c_uint8 * 8),
        ("VirtualSize", ctypes.c_uint32),
        ("VirtualAddress", ctypes.c_uint32),
        ("SizeOfRawData", ctypes.c_uint32),
        ("PointerToRawData", ctypes.c_uint32),
        ("PointerToRelocations", ctypes.c_uint32),
        ("PointerToLinenumbers", ctypes.c_uint32),
        ("NumberOfRelocations", ctypes.c_uint16),
        ("NumberOfLinenumbers", ctypes.c_uint16),
        ("Characteristics", ctypes.c_uint32),
    ]


class IMAGE_SYM_CLASS(IntEnum):
    """
    Symbol Storage Class
    """

    EXTERNAL = 2
    STATIC = 3
    LABEL = 6
    FUNCTION = 101


class CoffSymbolTableEntry(ctypes.LittleEndianStructure):
    """
    COFF Symbol Table Entry
    """

    _pack_ = 1
    _fields_ = [
        ("Name", ctypes.c_uint8 * 8),
        ("Value", ctypes.c_uint32),
        ("SectionNumber", ctypes.c_int16),
        ("Type", ctypes.c_uint16),
        ("StorageClass", ctypes.c_uint8),
        ("NumberOfAuxSymbols", ctypes.c_uint8),
    ]


class IMAGE_REL_I386(IntEnum):
    """
    i386 Relocation Types
    """

    DIR32 = 0x0006
    DIR32NB = 0x0007
    REL32 = 0x0014
    SECTION = 0x000A
    SECREL = 0x000B


class IMAGE_REL_AMD64(IntEnum):
    """
    AMD64 Relocation Types
    """

    ADDR64 = 0x0001
    ADDR32NB = 0x0003
    REL32 = 0x0004
    SECTION = 0x000A
    SECREL = 0x000B


class CoffRelocationTableEntry(ctypes.LittleEndianStructure):
    """
    COFF Relocations
    """

    _pack_ = 1
    _fields_ = [
        ("VirtualAddress", ctypes.c_uint32),
        ("SymbolTableIndex", ctypes.c_uint32),
        ("Type", ctypes.c_uint16),
    ]


class CoffParser:
    """
    Parses COFF object files.
    """

    data: bytes
    header: CoffFileHeader
    sections: list[CoffSectionTableEntry]
    relocations: list[list[CoffRelocationTableEntry]]
    symbols: list[CoffSymbolTableEntry]

    # Note: Symbols are uniquely identified by their index. It is possible for multiple symbols to have the same name so
    # in idx_to_symbol_name and symbol_name_to_idx, numeric suffixes are appended when necessary. To get the true name
    # of a symbol at index `symbol_idx`, call get_symbol_name(symbol_idx, true_name=True).
    idx_to_symbol_name: dict[int, str]
    symbol_name_to_idx: dict[str, int]

    def __init__(self, data: bytes):
        if data.startswith(b"\x00\x00\xff\xff"):
            raise ValueError(
                "This object file appears to have been compiled with whole program optimization (/GL flag)"
                " and cannot be parsed by this library"
            )
        self.data: bytes = data
        self._parse()

    def _parse(self) -> None:
        self.header = CoffFileHeader.from_buffer_copy(self.data)
        if self.header.Machine not in {
            IMAGE_FILE_MACHINE.I386,
            IMAGE_FILE_MACHINE.AMD64,
        }:
            raise NotImplementedError("Unsupported machine type")

        strings_offset = (
            self.header.PointerToSymbolTable + ctypes.sizeof(CoffSymbolTableEntry) * self.header.NumberOfSymbols
        )
        strings_size = struct.unpack("<I", self.data[strings_offset : strings_offset + 4])[0]
        self.strings: bytes = self.data[strings_offset : strings_offset + strings_size]

        self.symbols = []
        self.symbol_name_to_idx = {}
        self.idx_to_symbol_name = {}

        offset = self.header.PointerToSymbolTable
        aux = 0
        for i in range(self.header.NumberOfSymbols):
            symbol = CoffSymbolTableEntry.from_buffer_copy(self.data, offset)
            offset += ctypes.sizeof(CoffSymbolTableEntry)
            self.symbols.append(symbol)
            if aux:
                aux -= 1
                continue
            idx = len(self.symbols) - 1
            name = self.get_symbol_name(idx)
            aux = symbol.NumberOfAuxSymbols

            # Ensure unique symbol names
            i = 1
            base_name = name
            while name in self.symbol_name_to_idx:
                name = base_name + f"__{i}"
                i += 1
            self.symbol_name_to_idx[name] = idx
            self.idx_to_symbol_name[idx] = name

        self.sections = []
        self.relocations = []

        for i in range(self.header.NumberOfSections):
            offset = ctypes.sizeof(self.header) + ctypes.sizeof(CoffSectionTableEntry) * i
            section = CoffSectionTableEntry.from_buffer_copy(self.data, offset)
            self.sections.append(section)

            # Relocations
            relocs = []
            offset = section.PointerToRelocations
            for i in range(section.NumberOfRelocations):
                reloc = CoffRelocationTableEntry.from_buffer_copy(self.data, offset)
                relocs.append(reloc)
                offset += ctypes.sizeof(reloc)
            self.relocations.append(relocs)

    def get_symbol_name(self, symbol_idx: int, true_name: bool = False) -> str:
        if symbol_idx in self.idx_to_symbol_name and not true_name:
            return self.idx_to_symbol_name[symbol_idx]

        name_encoded = bytes(self.symbols[symbol_idx].Name)
        name_as_dwords = struct.unpack("<II", name_encoded)
        if name_as_dwords[0] == 0:
            name_encoded = extract_null_terminated_bytestr(self.strings, offset=name_as_dwords[1])
        return name_encoded.rstrip(b"\x00").decode("ascii")

    def get_section_name(self, section_idx: int) -> str:
        name = bytes(self.sections[section_idx].Name).rstrip(b"\x00").decode("ascii")
        if name.startswith("/"):
            return self.get_symbol_name(int(name[1:]))
        return name


class CoffSection(Section):
    """
    Section of the COFF object.
    """

    def __init__(
        self,
        name: str,
        file_offset: int,
        file_size: int,
        virtual_addr: int,
        virtual_size: int,
        coff_sec: CoffSectionTableEntry,
    ):
        super().__init__(name, file_offset, virtual_addr, virtual_size)
        self.filesize = file_size
        self._coff_sec = coff_sec

    @property
    def is_readable(self):
        return (self._coff_sec.Characteristics & IMAGE_SCN.MEM_READ) != 0

    @property
    def is_writable(self):
        return (self._coff_sec.Characteristics & IMAGE_SCN.MEM_WRITE) != 0

    @property
    def is_executable(self):
        return (self._coff_sec.Characteristics & IMAGE_SCN.MEM_EXECUTE) != 0

    @property
    def only_contains_uninitialized_data(self):
        return (self._coff_sec.Characteristics & IMAGE_SCN.CNT_UNINITIALIZED_DATA) != 0


class CoffRelocation(Relocation):
    """
    Relocation for a COFF object.
    """

    PACK_FORMAT = "<i"

    def relocate(self):
        value = self.value
        if value is None:
            log.debug("Unresolved relocation with no symbol.")
            return False
        self.owner.memory.store(self.relative_addr, struct.pack(self.PACK_FORMAT, value))
        return True


class CoffRelocationREL32(CoffRelocation):
    """
    Relocation for IMAGE_REL_*_REL32
    """

    @property
    def value(self):
        assert self.resolvedby is not None
        org_bytes = self.owner.memory.load(self.relative_addr, 4)
        org_value = struct.unpack("<I", org_bytes)[0]
        return org_value + self.resolvedby.rebased_addr - (self.rebased_addr + 4)


class CoffRelocationDIR32(CoffRelocation):
    """
    Relocation for IMAGE_REL_*_DIR32
    """

    @property
    def value(self):
        assert self.resolvedby is not None
        org_bytes = self.owner.memory.load(self.relative_addr, 4)
        org_value = struct.unpack("<I", org_bytes)[0]
        return org_value + self.resolvedby.rebased_addr


class CoffRelocationDIR32NB(CoffRelocation):
    """
    Relocation for IMAGE_REL_*_DIR32
    """

    @property
    def value(self):
        assert self.resolvedby is not None
        org_bytes = self.owner.memory.load(self.relative_addr, 4)
        org_value = struct.unpack("<I", org_bytes)[0]
        return org_value + self.resolvedby.relative_addr


class CoffRelocationADDR32NB(CoffRelocation):
    """
    Relocation for IMAGE_REL_AMD64_ADDR32NB
    """

    PACK_FORMAT = "<I"

    @property
    def value(self) -> int:
        assert self.resolvedby is not None
        return self.resolvedby.relative_addr


class CoffRelocationADDR64(CoffRelocation):
    """
    Relocation for IMAGE_REL_AMD64_ADDR64
    """

    PACK_FORMAT = "<Q"

    @property
    def value(self):
        assert self.resolvedby is not None
        return self.resolvedby.rebased_addr


class CoffRelocationSECTION(CoffRelocation):
    """
    Relocation for IMAGE_REL_*_SECTION
    """

    PACK_FORMAT = "<H"

    @property
    def value(self):
        assert isinstance(self.owner, Coff)
        section_idx = 0  # FIXME
        return section_idx


class CoffRelocationSECREL(CoffRelocation):
    """
    Relocation for IMAGE_REL_*_SECREL
    """

    PACK_FORMAT = "<I"

    @property
    def value(self):
        assert isinstance(self.owner, Coff)
        offset_to_symbol = 0  # FIXME
        return offset_to_symbol


RELOC_CLASSES: dict[IntEnum, dict[IntEnum, type[Relocation]]] = {
    IMAGE_FILE_MACHINE.I386: {
        IMAGE_REL_I386.REL32: CoffRelocationREL32,
        IMAGE_REL_I386.DIR32: CoffRelocationDIR32,
        IMAGE_REL_I386.DIR32NB: CoffRelocationDIR32NB,
        IMAGE_REL_I386.SECTION: CoffRelocationSECTION,
        IMAGE_REL_I386.SECREL: CoffRelocationSECREL,
    },
    IMAGE_FILE_MACHINE.AMD64: {
        IMAGE_REL_AMD64.ADDR64: CoffRelocationADDR64,
        IMAGE_REL_AMD64.ADDR32NB: CoffRelocationADDR32NB,
        IMAGE_REL_AMD64.REL32: CoffRelocationREL32,
        IMAGE_REL_AMD64.SECTION: CoffRelocationSECTION,
        IMAGE_REL_AMD64.SECREL: CoffRelocationSECREL,
    },
}

COFF_MACHINE_TO_ARCH_NAME = {
    IMAGE_FILE_MACHINE.I386: "x86",
    IMAGE_FILE_MACHINE.AMD64: "AMD64",
}


class Coff(Backend):
    """
    COFF object loader.
    """

    is_default = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.binary is None:
            self._data = self._binary_stream.read()
        else:
            with open(self.binary, "rb") as f:
                self._data = f.read()

        self._coff = CoffParser(self._data)

        arch = archinfo.arch_from_id(COFF_MACHINE_TO_ARCH_NAME[self._coff.header.Machine])
        self.set_arch(arch)

        # FIXME: Currently we just map the whole object file for convenience. Create a better memory map, discard object
        # file structure data.
        self._image_vmem = self._data

        # Add each section
        for section_idx, section in enumerate(self._coff.sections):
            section_name = self._coff.get_section_name(section_idx)
            vaddr = section.PointerToRawData
            vsize = section.SizeOfRawData
            self.segments.append(Segment(section.PointerToRawData, vaddr, section.SizeOfRawData, vsize))
            self.sections.append(
                CoffSection(
                    section_name,
                    section.PointerToRawData,
                    section.SizeOfRawData,
                    vaddr,
                    vsize,
                    section,
                )
            )

        self.memory.add_backer(0, bytes(self._image_vmem))
        self.mapped_base = self.linked_base = 0
        self.pic = True
        # assume windows, this can be wrong, but is more often right.
        self.os = "windows"
        self.supports_nx = False

        self._add_defined_symbols()
        self._add_relocs()

        # FIXME: Expose __imp_* symbols through self.imports

    def _add_defined_symbols(self) -> None:
        for sym_name, sym_idx in self._coff.symbol_name_to_idx.items():
            sym = self._coff.symbols[sym_idx]
            if sym.SectionNumber > 0 and sym.StorageClass in {
                IMAGE_SYM_CLASS.STATIC,
                IMAGE_SYM_CLASS.LABEL,
                IMAGE_SYM_CLASS.EXTERNAL,
            }:
                self.symbols.add(self.get_symbol(sym_name))

    def _add_relocs(self) -> None:
        for section_idx, section in enumerate(self._coff.sections):
            for reloc in self._coff.relocations[section_idx]:
                sym = self._coff.symbols[reloc.SymbolTableIndex]
                sym_name = self._coff.get_symbol_name(reloc.SymbolTableIndex)
                patch_offset = section.PointerToRawData + reloc.VirtualAddress

                if sym.StorageClass in {
                    IMAGE_SYM_CLASS.STATIC,
                    IMAGE_SYM_CLASS.LABEL,
                    IMAGE_SYM_CLASS.EXTERNAL,
                }:
                    reloc_class = RELOC_CLASSES[self._coff.header.Machine].get(reloc.Type, None)
                    if reloc_class is not None:
                        cle_symbol = self.get_symbol(sym_name, produce_extern_symbols=True)
                        self.relocs.append(reloc_class(self, cle_symbol, patch_offset))
                        continue

                log.warning("Skipped relocation type %#x at %#x for symbol %s", reloc.Type, patch_offset, sym_name)

    @classmethod
    def is_compatible(cls, stream):
        stream.seek(0)
        identstring = stream.read(2)
        stream.seek(0)
        return int.from_bytes(identstring, "little") in (IMAGE_FILE_MACHINE.I386, IMAGE_FILE_MACHINE.AMD64)

    def get_symbol(self, name: str, produce_extern_symbols: bool = False) -> Symbol | None:
        if name not in self._coff.symbol_name_to_idx:
            return None

        if name == "__ImageBase":
            return Symbol(self, name, 0, 0, SymbolType.TYPE_OTHER)

        sym = self._coff.symbols[self._coff.symbol_name_to_idx[name]]
        if sym.StorageClass in {
            IMAGE_SYM_CLASS.STATIC,
            IMAGE_SYM_CLASS.LABEL,
            IMAGE_SYM_CLASS.EXTERNAL,
        }:
            symbol_type = SymbolType.TYPE_FUNCTION if sym.Type == 0x20 else SymbolType.TYPE_OTHER
            if sym.SectionNumber > 0:
                sym_addr = self._coff.sections[sym.SectionNumber - 1].PointerToRawData + sym.Value
                return Symbol(self, name, sym_addr, 1, symbol_type)
            elif sym.SectionNumber == 0:
                if produce_extern_symbols:
                    return Symbol(self, name, 0, sym.Value, symbol_type)
                return None

        raise NotImplementedError("Unsupported symbol")


register_backend("COFF", Coff)



from __future__ import annotations

import logging
import os
import struct
from collections import defaultdict

import elftools

from cle.address_translator import AT
from cle.backends import register_backend
from cle.backends.blob import Blob
from cle.backends.region import Segment
from cle.errors import CLECompatibilityError, CLEError
from cle.memory import Clemory

from .elf import ELF

log = logging.getLogger(name=__name__)

# TODO: yall know struct.unpack_from exists, right? maybe even bitstream?


class ELFCore(ELF):
    """
    Loader class for ELF core files.

    One key pain point when analyzing a core dump generated on a remote machine is that the paths to binaries are
    absolute (and may not exist or be the same on your local machine).

    Therefore, you can use the options ```remote_file_mapping`` to specify a ``dict`` mapping (easy if there are a small
    number of mappings) or ``remote_file_mapper`` to specify a function that accepts a remote file name and returns the
    local file name (useful if there are many mappings).

    If you specify both ``remote_file_mapping`` and ``remote_file_mapper``, ``remote_file_mapping`` is applied first,
    then the result is passed to ``remote_file_mapper``.

    :param executable:           Optional path to the main binary of the core dump. If not supplied, ELFCore will
                                 attempt to figure it out automatically from the core dump.
    :param remote_file_mapping:  Optional dict that maps specific file names in the core dump to other file names.
    :param remote_file_mapper:   Optional function that is used to map every file name in the core dump to whatever is
                                 returned from this function.
    """

    is_default = True  # Tell CLE to automatically consider using the ELFCore backend

    def __init__(self, *args, executable=None, remote_file_mapping=None, remote_file_mapper=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.filename_lookup = []
        self.__current_thread = None
        self._threads = []
        self.auxv = {}
        self.pr_fname = None
        self._main_filepath = executable
        self._page_size = 0x1000  # a default page size, will be changed later by parsing notes
        self._main_object = None

        if remote_file_mapping is not None:
            self._remote_file_mapper = lambda x: remote_file_mapping.get(x, x)
        else:
            self._remote_file_mapper = lambda x: x

        if remote_file_mapper is not None:
            orig = self._remote_file_mapper
            self._remote_file_mapper = lambda x: remote_file_mapper(orig(x))

        self.__extract_note_info()

        self.__reload_children()

        self._remote_file_mapper = None

    @staticmethod
    def is_compatible(stream):
        stream.seek(0)
        identstring = stream.read(0x1000)
        stream.seek(0)
        if identstring.startswith(b"\x7fELF"):
            if elftools.elf.elffile.ELFFile(stream).header["e_type"] == "ET_CORE":
                return True
            return False
        return False

    def __cycle_thread(self):
        if self.__current_thread is not None:
            self._threads.append(self.__current_thread)
        self.__current_thread = {}

    @property
    def threads(self):
        return list(range(len(self._threads)))

    def thread_registers(self, thread=None):
        if thread is None:
            thread = 0
        return self._threads[thread]["registers"]

    def __extract_note_info(self):
        """
        All meaningful information about the process's state at crashtime is stored in the note segment.
        """
        for seg_readelf in self._reader.iter_segments():
            if seg_readelf.header.p_type == "PT_NOTE":
                for note in seg_readelf.iter_notes():
                    if note.n_type == "NT_PRSTATUS":
                        self.__cycle_thread()
                        n_desc = note.n_desc.encode("latin-1") if isinstance(note.n_desc, str) else note.n_desc
                        self.__parse_prstatus(n_desc)
                    elif note.n_type == "NT_PRPSINFO":
                        self.__parse_prpsinfo(note.n_desc)
                    elif note.n_type == "NT_AUXV":
                        n_desc = note.n_desc.encode("latin-1") if isinstance(note.n_desc, str) else note.n_desc
                        self.__parse_auxv(n_desc)
                    elif note.n_type == "NT_FILE":
                        self.__parse_files(note.n_desc)
                    elif note.n_type == 512 and self.arch.name == "X86":
                        n_desc = note.n_desc.encode("latin-1") if isinstance(note.n_desc, str) else note.n_desc
                        self.__parse_x86_tls(n_desc)

        self._replace_main_object_path()

        self.__cycle_thread()
        if not self._threads:
            log.warning("Could not find thread info, cannot initialize registers")
        elif self.arch.name == "X86" and "segments" not in self._threads[0]:
            if "AT_RANDOM" in self.auxv:
                log.warning(
                    "This core dump does not contain TLS information. "
                    "threads will be matched to TLS regions via heuristics"
                )
                pointer_rand = self.auxv["AT_RANDOM"][4:8]
                all_locations = [
                    addr - 0x18
                    for addr in self.__dummy_clemory.find(pointer_rand)
                    if self.__dummy_clemory.unpack_word(addr - 0x18) == addr - 0x18
                ]
                # the heuristic is that generally threads are allocated with descending tls addresses
                for thread, loc in zip(self._threads, reversed(all_locations)):
                    thread["segments"] = {thread["registers"]["gs"] >> 3: (loc, 0xFFFFF, 0x51)}
            else:
                log.warning("This core dump does not contain TLS or auxv information. TLS information will be wrong.")
                for thread in self._threads:
                    thread["segments"] = {thread["registers"]["gs"] >> 3: (0, 0xFFFFFFFF, 0x51)}

    def _replace_main_object_path(self):
        """
        try to replace path of the main_object with the specified one
        """
        if not self._main_filepath or not self.filename_lookup:
            return

        # identify the original path and assuming pr_fname always exists
        matched = None
        for i, (a, b, c, fn) in enumerate(self.filename_lookup):
            if os.path.basename(fn).startswith(
                self.pr_fname
            ):  # pr_fname is defined to be the first 16 bytes of the executable name
                matched = fn
                break
        else:
            raise CLEError("Fail to find the main object, is this core dump malformed?")

        # replace the path
        for i, (a, b, c, fn) in enumerate(self.filename_lookup):
            if fn == matched:
                self.filename_lookup[i] = (a, b, c, self._main_filepath)

    @property
    def __dummy_clemory(self):
        dummy_clemory = Clemory(self.arch, root=True)
        dummy_clemory.add_backer(self.linked_base, self.memory)
        return dummy_clemory

    def __parse_prstatus(self, desc):
        """
        Parse out the prstatus, accumulating the general purpose register values.
        Supports AMD64, X86, ARM, AArch64, MIPS and MIPSEL at the moment.

        :param prstatus: a note object of type NT_PRSTATUS.
        """

        # TODO: support all architectures angr supports
        arch_bytes = self.arch.bytes
        if arch_bytes == 4:
            fmt = "I"
        elif arch_bytes == 8:
            fmt = "Q"
        else:
            raise CLEError("Architecture must have a bitwidth of either 64 or 32")

        end = ">" if self.arch.memory_endness == "Iend_BE" else "<"

        pos = 0

        def read_longs(n):
            fin = pos + n * arch_bytes
            return (fin, *struct.unpack(end + fmt * n, desc[pos:fin]))

        def read_ints(n):
            fin = pos + n * 4
            return (fin, *struct.unpack(end + "I" * n, desc[pos:fin]))

        def read_timeval():
            sec, usec = struct.unpack(end + fmt * 2, desc[pos : pos + 2 * arch_bytes])
            return (pos + 2 * arch_bytes, sec * 1000000 + usec)

        result = {}

        pos, result["si_signo"], result["si_code"], result["si_errno"] = read_ints(3)

        # this field is a short, but it's padded to an int
        (result["pr_cursig"],) = struct.unpack(end + "H", desc[pos : pos + 2])
        pos += 4

        pos, result["pr_sigpend"], result["pr_sighold"] = read_longs(2)

        pos, result["pr_pid"], result["pr_ppid"], result["pr_pgrp"], result["pr_sid"] = read_ints(4)

        pos, result["pr_utime_usec"] = read_timeval()
        pos, result["pr_stime_usec"] = read_timeval()
        pos, result["pr_cutime_usec"] = read_timeval()
        pos, result["pr_cstime_usec"] = read_timeval()

        # parse out general purpose registers
        if self.arch.name == "AMD64":
            # register names as they appear in dump
            rnames = [
                "r15",
                "r14",
                "r13",
                "r12",
                "rbp",
                "rbx",
                "r11",
                "r10",
                "r9",
                "r8",
                "rax",
                "rcx",
                "rdx",
                "rsi",
                "rdi",
                "xxx",
                "rip",
                "cs",
                "eflags",
                "rsp",
                "ss",
                "fs_base",
                "gs_base",
                "ds",
                "es",
                "xxx",
                "xxx",
            ]
            nreg = 27
        elif self.arch.name == "X86":
            rnames = [
                "ebx",
                "ecx",
                "edx",
                "esi",
                "edi",
                "ebp",
                "eax",
                "ds",
                "es",
                "fs",
                "gs",
                "xxx",
                "eip",
                "cs",
                "eflags",
                "esp",
                "ss",
            ]
            nreg = 17
        elif self.arch.name == "ARMHF" or self.arch.name == "ARMEL":
            rnames = [
                "r0",
                "r1",
                "r2",
                "r3",
                "r4",
                "r5",
                "r6",
                "r7",
                "r8",
                "r9",
                "r10",
                "r11",
                "r12",
                "r13",
                "r14",
                "r15",
                "xxx",
                "xxx",
            ]
            nreg = 18
        elif self.arch.name == "AARCH64":
            rnames = [f"x{i}" for i in range(32)]
            rnames.append("pc")
            rnames.append("xxx")
            nreg = 34
        elif self.arch.name == "MIPS32":
            rnames = [
                "xxx",
                "xxx",
                "xxx",
                "xxx",
                "xxx",
                "xxx",
                "zero",
                "at",
                "v0",
                "v1",
                "a0",
                "a1",
                "a2",
                "a3",
                "t0",
                "t1",
                "t2",
                "t3",
                "t4",
                "t5",
                "t6",
                "t7",
                "s0",
                "s1",
                "s2",
                "s3",
                "s4",
                "s5",
                "s6",
                "s7",
                "t8",
                "t9",
                "k0",
                "k1",
                "gp",
                "sp",
                "s8",
                "ra",
                "lo",
                "hi",
                "pc",
                "bad",
                "sr",
                "status",
                "cause",
            ]
            nreg = 45
        else:
            raise CLECompatibilityError(f"Architecture '{self.arch.name}' unsupported by ELFCore")

        assert nreg == len(rnames), "Please create an issue with this core-file attached to get this fixed."
        pos, *regvals = read_longs(nreg)
        result["registers"] = dict(zip(rnames, regvals))
        del result["registers"]["xxx"]

        pos, result["pr_fpvalid"] = read_ints(1)
        assert (
            pos <= len(desc) < pos + arch_bytes
        ), "Please create an issue with this core-file attached to get this fixed."

        self.__current_thread.update(result)

    def __parse_prpsinfo(self, desc):
        pr_fname = desc.pr_fname.split(b"\x00", 1)[0]
        try:
            self.pr_fname = pr_fname.decode()
        except UnicodeDecodeError:
            self.pr_fname = repr(pr_fname)

    def __parse_files(self, desc):
        self._page_size = desc.page_size
        self.filename_lookup = [
            (ent.vm_start, ent.vm_end, ent.page_offset * desc.page_size, self._remote_file_mapper(fn.decode()))
            for ent, fn in zip(desc.Elf_Nt_File_Entry, desc.filename)
        ]

    def __parse_x86_tls(self, desc):
        self.__current_thread["segments"] = {}
        for offset in range(0, len(desc), 4 * 4):
            index, base, limit, flags = struct.unpack_from("4I", desc, offset)
            self.__current_thread["segments"][index] = (base, limit, flags)

    def __parse_auxv(self, desc):
        for offset in range(0, len(desc), self.arch.bytes * 2):
            code = struct.unpack_from(self.arch.struct_fmt(), desc, offset)[0]
            value = struct.unpack_from(self.arch.struct_fmt(), desc, offset + self.arch.bytes)[0]
            code_str = auxv_codes.get(code, code)

            if code_str == "AT_RANDOM":
                value = self.__dummy_clemory.load(value, 0x10)
            elif code_str in ("AT_EXECFN", "AT_PLATFORM"):
                pos = value
                value = bytearray()
                while True:
                    byte = self.__dummy_clemory[pos]
                    if byte == 0:
                        break
                    value.append(byte)
                    pos += 1
                value = bytes(value)

            self.auxv[code_str] = value

    def __reload_children(self):
        self.loader.page_size = self._page_size
        self.loader._perform_relocations = False

        # hack: we are using a loader internal method in a non-kosher way which will cause our children to be
        # marked as the main binary if we are also the main binary
        # work around this by setting ourself here:
        if self.loader._main_object is None:
            self.loader._main_object = self

        child_patches = defaultdict(list)
        for vm_start, vm_end, offset, filename in self.filename_lookup:
            try:
                patch_data = self.__dummy_clemory.load(vm_start, vm_end - vm_start)
            except KeyError:
                pass
            else:
                child_patches[filename].append((vm_start, offset, patch_data))

        remaining_segments = list(self.segments)

        for filename, patches in child_patches.items():
            try:
                with open(filename, "rb") as fp:
                    obj = self.loader._load_object_isolated(fp)
            except (FileNotFoundError, PermissionError, CLECompatibilityError) as ex:
                if isinstance(ex, FileNotFoundError):
                    log.warning(
                        "Dependency %s does not exist on the current system; this core may be incomplete.", filename
                    )
                elif isinstance(ex, CLECompatibilityError):
                    log.warning("Could not find a compatible loader for %s; this core may be incomplete.", filename)
                else:
                    log.warning("Could not load %s; this core may be incomplete.", filename)
                if self.loader._main_object is self:
                    self.loader._main_object = None
                self.child_objects.clear()
                return

            # several ways to try to match the NT_FILE entries to the object
            # (not trivial because offsets can be mapped multiple places)
            # (and because there's no clear pattern for how mappings are included or omitted)
            base_addr = None

            # try one: use the delta between each allocation as a signature (works when the text segment is missing)
            if base_addr is None:
                vm_starts = [a for a, _, _ in patches]
                vm_deltas = [b - a for a, b in zip(vm_starts, vm_starts[1:])]
                segment_starts = [seg.vaddr for seg in obj.segments]
                segment_deltas = [b - a for a, b in zip(segment_starts, segment_starts[1:])]

                # funky lil algorithm to find substrings
                for match_idx in range(len(segment_deltas) - len(vm_deltas) + 1):
                    for idx, vm_delta in enumerate(vm_deltas):
                        if vm_delta != segment_deltas[match_idx + idx]:
                            break
                    else:
                        base_addr = vm_starts[0] - AT.from_lva(obj.segments[match_idx].vaddr, obj).to_rva()
                        break

            # try two: if the file is identity-mapped, it's easy (?)
            if base_addr is None:
                base_reccomendations = [a - b for a, b, _ in patches]
                if all(a == base_reccomendations[0] for a in base_reccomendations):
                    base_addr = base_reccomendations[0]

            # try three: if we have the zero offset then it's easy (?)
            if base_addr is None:
                if patches[0][1] == 0:
                    base_addr = patches[0][0]

            if base_addr is None:
                log.warning("Could not load %s (could not determine base); core may be incomplete", filename)
                if self.loader._main_object is self:
                    self.loader._main_object = None
                self.child_objects.clear()
                return

            obj._custom_base_addr = base_addr
            self.child_objects.append(obj)

            # figure out how the core's data should affect the child object's data
            # iterate over all the core segments, since the only time we will need to make a change to the child's
            # memory is if the core has something to say about it
            # if there is ANY OVERLAP AT ALL, copy over the relevant data and nuke the segment
            # then, if there is any part of the segment which DOESN'T correspond to a child segment, inject a new memory
            #  backer into the child for the relevant data

            max_addr = base_addr + (obj.max_addr - obj.min_addr)
            i = 0
            while i < len(remaining_segments):
                seg = remaining_segments[i]
                # check for overlap (overapproximation)
                if base_addr <= seg.vaddr <= max_addr or seg.vaddr <= base_addr < seg.vaddr + seg.memsize:
                    remaining_segments.pop(i)

                    # if there is data before the beginning of the child or after the end,
                    # make new artificial segments for it
                    if seg.vaddr < base_addr:
                        size = base_addr - seg.vaddr
                        remaining_segments.insert(i, Segment(seg.offset, seg.vaddr, size, size))
                        i += 1
                    if seg.max_addr > max_addr:
                        size = seg.max_addr - max_addr
                        offset = seg.memsize - size
                        remaining_segments.insert(i, Segment(seg.offset + offset, seg.vaddr + offset, size, size))
                        i += 1

                    # ohhhh this is SUCH a confusing address space-conversation problem!
                    # we're going to enumerate the contents of the core segment. at each point we find the relevant
                    # child backer. if this skips any content, inject a backer into the child.
                    # then, copy the contents of the core segment that overlaps the child backer.
                    cursor = max(0, base_addr - seg.vaddr)
                    # use filesize and not memsize so we don't overwrite stuff with zeroes if it's omitted from the core
                    while cursor < seg.filesize:
                        child_cursor = cursor + seg.vaddr - base_addr
                        try:
                            child_offset, child_backer = next(obj.memory.backers(child_cursor))
                        except StopIteration:
                            # is this right? is there any behavior we need to account for in the case that there is
                            # somehow no backer past a point mapped by the core?
                            break

                        # have we skipped any part of the core?
                        skip_size = child_offset - child_cursor
                        if skip_size > 0:
                            # inject it into the child
                            obj.memory.add_backer(
                                child_cursor,
                                self.memory.load(AT.from_mva(cursor + seg.vaddr, self).to_rva(), skip_size),
                            )

                        # how much of the child's segment have we skipped by
                        # starting at the beginning of the core segment?
                        child_backer_offset = max(0, -skip_size)
                        # how much of the core's segment have we skipped and handled via injection?
                        core_backer_offset = max(0, skip_size)
                        # how much can we copy?
                        copy_size = min(
                            len(child_backer) - child_backer_offset, seg.memsize - (cursor + core_backer_offset)
                        )
                        if copy_size > 0:
                            # do the copy if we have anything to copy
                            obj.memory.store(
                                child_offset + child_backer_offset,
                                self.memory.load(
                                    AT.from_mva(seg.vaddr + cursor + core_backer_offset, self).to_rva(), copy_size
                                ),
                            )

                        # advance cursor
                        cursor += core_backer_offset + copy_size
                else:
                    i += 1

        # for all remaining segments, make blobs out of them
        mem = self.__dummy_clemory
        for seg in remaining_segments:
            if not seg.memsize:
                continue
            obj = Blob(
                self.binary,
                mem,
                segments=[(seg.vaddr, seg.vaddr, seg.memsize)],
                base_addr=seg.vaddr,
                arch=self.arch,
                entry_point=0,
                force_rebase=True,
            )
            self.child_objects.append(obj)

        self.mapped_base = 0
        self._max_addr = 0
        self.has_memory = False
        if self.loader._main_object is self:
            self.loader._main_object = None
            self.__record_main_object()

    def __record_main_object(self):
        """
        If children objects are reloaded, identify the main object for later use by loader
        """
        for obj in self.child_objects:
            if self.pr_fname and obj.binary_basename.startswith(self.pr_fname):
                self._main_object = obj
                return
            if self._main_filepath is not None and os.path.basename(self._main_filepath) == obj.binary_basename:
                self._main_object = obj
                return

        log.warning("Failed to identify main object in ELFCore")
        self._main_object = self


auxv_codes = {
    0x0: "AT_NULL",
    0x1: "AT_IGNORE",
    0x2: "AT_EXECFD",
    0x3: "AT_PHDR",
    0x4: "AT_PHENT",
    0x5: "AT_PHNUM",
    0x6: "AT_PAGESZ",
    0x7: "AT_BASE",
    0x8: "AT_FLAGS",
    0x9: "AT_ENTRY",
    0xA: "AT_NOTELF",
    0xB: "AT_UID",
    0xC: "AT_EUID",
    0xD: "AT_GID",
    0xE: "AT_EGID",
    0x11: "AT_CLKTCK",
    0xF: "AT_PLATFORM",
    0x10: "AT_HWCAP",
    0x12: "AT_FPUCW",
    0x13: "AT_DCACHEBSIZE",
    0x14: "AT_ICACHEBSIZE",
    0x15: "AT_UCACHEBSIZE",
    0x16: "AT_IGNOREPPC",
    0x17: "AT_SECURE",
    0x18: "AT_BASE_PLATFORM",
    0x19: "AT_RANDOM",
    0x1A: "AT_HWCAP2",
    0x1F: "AT_EXECFN",
    0x20: "AT_SYSINFO",
    0x21: "AT_SYSINFO_EHDR",
    0x22: "AT_L1I_CACHESHAPE",
    0x23: "AT_L1D_CACHESHAPE",
    0x24: "AT_L2_CACHESHAPE",
    0x25: "AT_L3_CACHESHAPE",
    0x28: "AT_L1I_CACHESIZE",
    0x29: "AT_L1I_CACHEGEOMETRY",
    0x2A: "AT_L1D_CACHESIZE",
    0x2B: "AT_L1D_CACHEGEOMETRY",
    0x2C: "AT_L2_CACHESIZE",
    0x2D: "AT_L2_CACHEGEOMETRY",
    0x2E: "AT_L3_CACHESIZE",
    0x2F: "AT_L3_CACHEGEOMETRY",
}

register_backend("elfcore", ELFCore)



from __future__ import annotations

from cle.backends.region import Section


class PESection(Section):
    """
    Represents a section for the PE format.
    """

    def __init__(self, pe_section, remap_offset=0, name: str | None = None):
        super().__init__(
            name or pe_section.Name.decode("latin-1"),  # ensure all bytes can be decoded
            pe_section.PointerToRawData,
            pe_section.VirtualAddress + remap_offset,
            pe_section.Misc_VirtualSize,
        )

        self.characteristics = pe_section.Characteristics
        self.filesize = pe_section.SizeOfRawData

    #
    # Public properties
    #

    @property
    def is_readable(self):
        return self.characteristics & 0x40000000 != 0

    @property
    def is_writable(self):
        return self.characteristics & 0x80000000 != 0

    @property
    def is_executable(self):
        return self.characteristics & 0x20000000 != 0

    @property
    def only_contains_uninitialized_data(self):
        return self.filesize == 0



from __future__ import annotations

import logging

from cle.errors import CLEError

from .backend import Backend, register_backend

try:
    import arpy
except ImportError:
    arpy = None

log = logging.getLogger(__name__)


class StaticArchive(Backend):
    @classmethod
    def is_compatible(cls, stream):
        stream.seek(0)
        return stream.read(8) == b"!<arch>\n"

    is_default = True
    is_outer = True

    def __init__(self, *args, **kwargs):
        if arpy is None:
            raise CLEError("run `pip install arpy==1.1.1` to load archive files")
        super().__init__(*args, **kwargs)

        # hack: we are using a loader internal method in a non-kosher way which will cause our children to be
        # marked as the main binary if we are also the main binary
        # work around this by setting ourself here:
        if self.loader._main_object is None:
            self.loader._main_object = self

        ar = arpy.Archive(fileobj=self._binary_stream)
        ar.read_all_headers()
        for name, stream in ar.archived_files.items():
            child = self.loader._load_object_isolated(stream)
            child.binary = child.binary_basename = name.decode()
            child.parent_object = self
            self.child_objects.append(child)

        if self.child_objects:
            self._arch = self.child_objects[0].arch
        else:
            log.warning("Loaded empty static archive?")
        self.has_memory = False
        self.pic = True

        # hack pt. 2
        if self.loader._main_object is self:
            self.loader._main_object = None


register_backend("AR", StaticArchive)



from __future__ import annotations

import logging

from cle.address_translator import AT

from .tls_object import InternalTLSRelocation, ThreadManager, TLSObject

_l = logging.getLogger(__name__)


class PEThreadManager(ThreadManager):
    def register_object(self, obj):
        if not super().register_object(obj):
            return False

        # The PE TLS header says to write its index into a given address
        if hasattr(obj, "tls_index_address"):
            rva = AT.from_lva(obj.tls_index_address, obj).to_rva()
            if rva > 0 and rva in obj.memory:
                try:
                    obj.memory.pack_word(rva, obj.tls_module_id)
                except KeyError:
                    _l.warning(
                        "TLS index address %#x (RVA %#x) does not exist or is not writable.", obj.tls_index_address, rva
                    )
            else:
                # apparently RVA can be an absolute address pointing to another object. we don't handle it yet.
                _l.warning(
                    "TLS index address %#x (RVA %#x) is not mapped in the current object.", obj.tls_index_address, rva
                )
        return True

    @property
    def _thread_cls(self):
        return PETLSObject


class PETLSObject(TLSObject):
    """
    This class is used when parsing the Thread Local Storage of a PE binary. It
    represents both the TLS array and the TLS data area for a specific thread.

    In memory the ``PETLSObj`` is laid out as follows::

        +----------------------+---------------------------------------+
        | TLS array            | TLS data area                         |
        +----------------------+---------------------------------------+

    A more detailed description of the TLS array and TLS data areas is given
    below.

    **TLS array**

    The TLS array is an array of addresses that points into the TLS data area.
    In memory it is laid out as follows::

        +-----------+-----------+-----+-----------+
        |  address  |  address  | ... |  address  |
        +-----------+-----------+-----+-----------+
        | index = 0 | index = 1 |     | index = n |
        +-----------+-----------+-----+-----------+

    The size of each address is architecture independent (e.g. on X86 it is
    4 bytes). The number of addresses in the TLS array is equal to the number
    of modules that contain TLS data. At load time (i.e. in the ``finalize``
    method), each module is assigned an index into the TLS array. The address
    of this module's TLS data area is then stored at this location in the
    array.

    **TLS data area**

    The TLS data area directly follows the TLS array and contains the actual
    TLS data for each module. In memory it is laid out as follows::

        +----------+-----------+----------+-----------+-----+
        | TLS data | zero fill | TLS data | zero fill | ... |
        +----------+-----------+----------+-----------+-----+
        |       module a       |       module b       | ... |
        +---------------------------------------------------+

    The size of each module's TLS data area is variable and can be found in the
    module's ``tls_data_size`` property. The same applies to the zero fill. At
    load time (i.e in the ``finalize`` method), the initial TLS data values are
    copied into the TLS data area. Because a TLS index is also assigned to each
    module, we can access a module's TLS data area using this index into the
    TLS array to get the start address of the TLS data.
    """

    def __init__(self, thread_manager: PEThreadManager):
        super().__init__(loader=thread_manager.loader, arch=thread_manager.arch)

        self.used_modules = len(thread_manager.modules)
        self.data_start = self.arch.bytes * thread_manager.max_modules
        self.used_data = 0
        self.memory.add_backer(0, bytes(self.data_start))
        self.pic = True

        for obj in thread_manager.modules:
            image = thread_manager.initialization_image(obj)
            if image is None:
                continue
            image_offset = self.data_start + self.used_data
            index_offset = obj.tls_module_id * self.arch.bytes

            self.memory.pack_word(index_offset, image_offset)
            self.relocs.append(InternalTLSRelocation(image_offset, index_offset, self))
            self.memory.add_backer(image_offset, image)
            self.used_data += len(image)

    def get_tls_data_addr(self, tls_idx):
        """
        Get the start address of a module's TLS data area via the module's TLS
        index.

        From the PE/COFF spec:

            The code uses the TLS index and the TLS array location (multiplying
            the index by the word size and using it as an offset into the
            array) to get the address of the TLS data area for the given
            program and module.
        """
        if 0 <= tls_idx < self.used_modules:
            return self.memory.unpack_word(tls_idx * self.arch.bytes)
        else:
            raise IndexError("TLS index out of range")

    @property
    def max_addr(self):
        return self.mapped_base + self.data_start + self.used_data - 1

    # PE is MUCH simpler in terms of what's the pointer to the thread data. Add these properties for compatibility.

    @property
    def thread_pointer(self):
        return self.mapped_base

    @property
    def user_thread_pointer(self):
        return self.mapped_base



from __future__ import annotations


class Region:
    """
    A region of memory that is mapped in the object's file.

    :ivar offset:       The offset into the file the region starts.
    :ivar vaddr:        The virtual address.
    :ivar filesize:     The size of the region in the file.
    :ivar memsize:      The size of the region when loaded into memory.

    The prefix `v-` on a variable or parameter name indicates that it refers to the virtual, loaded memory space,
    while a corresponding variable without the `v-` refers to the flat zero-based memory of the file.

    When used next to each other, `addr` and `offset` refer to virtual memory address and file offset, respectively.
    """

    vaddr: int
    memsize: int
    filesize: int

    def __init__(self, offset, vaddr, filesize, memsize):
        self.vaddr = vaddr
        self.memsize = memsize
        self.filesize = filesize
        self.offset = offset

    def _rebase(self, delta):
        """
        Does region rebasing to other base address.
        Intended for usage by loader's add_object to reflect the rebasing.

        :param int delta: Delta offset between an old and a new image bases
        """
        self.vaddr += delta

    def contains_addr(self, addr):
        """
        Does this region contain this virtual address?
        """
        return self.vaddr <= addr < self.vaddr + self.memsize

    def contains_offset(self, offset):
        """
        Does this region contain this offset into the file?
        """
        return self.offset <= offset < self.offset + self.filesize

    def addr_to_offset(self, addr):
        """
        Convert a virtual memory address into a file offset
        """
        offset = addr - self.vaddr + self.offset
        if not self.contains_offset(offset):
            return None
        return offset

    def offset_to_addr(self, offset):
        """
        Convert a file offset into a virtual memory address
        """
        addr = offset - self.offset + self.vaddr
        if not self.contains_addr(addr):
            return None
        return addr

    def __repr__(self):
        return "<{} {}>".format(
            self.__class__.__name__,
            ", ".join([f"{k}={hex(v) if v is not None else 'None'}" for k, v in self.__dict__.items()]),
        )

    @property
    def max_addr(self):
        """
        The maximum virtual address of this region
        """
        return self.vaddr + self.memsize - 1

    @property
    def min_addr(self):
        """
        The minimum virtual address of this region
        """
        return self.vaddr

    @property
    def max_offset(self):
        """
        The maximum file offset of this region
        """
        return self.offset + self.filesize - 1

    def min_offset(self):
        """
        The minimum file offset of this region
        """
        return self.offset

    # EDG says: Blobs now have segments, and SimOS will get upset if these don't exist.  See simos.py line 107 for
    # some code you should probably fix if you don't like it.
    @property
    def is_readable(self) -> bool:
        # pylint: disable=no-self-use
        return True

    @property
    def is_writable(self) -> bool:
        # pylint: disable=no-self-use
        return True

    @property
    def is_executable(self) -> bool:
        # pylint: disable=no-self-use
        return True


class Segment(Region):
    pass


class EmptySegment(Segment):
    """
    A segment with no static content, and permissions
    """

    def __init__(self, vaddr, memsize, is_readable=True, is_writable=True, is_executable=False):
        super().__init__(0, vaddr, 0, memsize)
        self._is_readable = is_readable
        self._is_writable = is_writable
        self._is_executable = is_executable

    @property
    def is_executable(self):
        return self._is_executable

    @property
    def is_writable(self):
        return self._is_writable

    @property
    def is_readable(self):
        return self._is_readable

    @property
    def only_contains_uninitialized_data(self):
        """
        Whether this section is initialized to zero after the executable is loaded.
        """
        return True


class Section(Region):
    """
    Simple representation of a loaded section.

    :ivar str name:     The name of the section
    """

    def __init__(self, name, offset, vaddr, size):
        """
        :param str name:    The name of the section
        :param int offset:  The offset into the binary file this section begins
        :param int vaddr:   The address in virtual memory this section begins
        :param int size:    How large this section is
        """
        super().__init__(offset, vaddr, size, size)
        self.name = name

    @property
    def is_readable(self):
        """
        Whether this section has read permissions
        """
        raise NotImplementedError()

    @property
    def is_writable(self):
        """
        Whether this section has write permissions
        """
        raise NotImplementedError()

    @property
    def is_executable(self):
        """
        Whether this section has execute permissions
        """
        raise NotImplementedError()

    @property
    def only_contains_uninitialized_data(self):
        """
        Whether this section is initialized to zero after the executable is loaded.
        """
        raise NotImplementedError()

    def __repr__(self):
        return "<{} | offset {:#x}, vaddr {:#x}, size {:#x}>".format(
            self.name if self.name else "Unnamed", self.offset, self.vaddr, self.memsize
        )



"""Relocations for amd64/x86_64

Reference: https://gitlab.com/x86-psABIs/x86-64-ABI/-/jobs/artifacts/master/raw/x86-64-ABI/abi.pdf?job=build page 73
"""

from __future__ import annotations

from .generic import (
    GenericAbsoluteAddendReloc,
    GenericCopyReloc,
    GenericIRelativeReloc,
    GenericJumpslotReloc,
    GenericPCRelativeAddendReloc,
    GenericRelativeReloc,
    GenericTLSDoffsetReloc,
    GenericTLSModIdReloc,
    GenericTLSOffsetReloc,
    RelocGOTMixin,
    RelocTruncate32Mixin,
)


class R_X86_64_64(GenericAbsoluteAddendReloc):
    pass


class R_X86_64_COPY(GenericCopyReloc):
    pass


class R_X86_64_RELATIVE(GenericRelativeReloc):
    pass


class R_X86_64_IRELATIVE(GenericIRelativeReloc):
    pass


class R_X86_64_GLOB_DAT(GenericJumpslotReloc):
    pass


class R_X86_64_JUMP_SLOT(GenericJumpslotReloc):
    pass


class R_X86_64_DTPMOD64(GenericTLSModIdReloc):
    pass


class R_X86_64_DTPOFF64(GenericTLSDoffsetReloc):
    pass


class R_X86_64_TPOFF64(GenericTLSOffsetReloc):
    pass


class R_X86_64_PC32(RelocTruncate32Mixin, GenericPCRelativeAddendReloc):
    check_sign_extend = True


class R_X86_64_32(RelocTruncate32Mixin, GenericAbsoluteAddendReloc):
    check_zero_extend = True


class R_X86_64_32S(RelocTruncate32Mixin, GenericAbsoluteAddendReloc):
    check_sign_extend = True


class R_X86_64_PLT32(RelocTruncate32Mixin, GenericPCRelativeAddendReloc):
    check_sign_extend = True


class R_X86_64_GOTPCREL(RelocGOTMixin, RelocTruncate32Mixin, GenericPCRelativeAddendReloc):
    check_sign_extend = True


class R_X86_64_GOTPCRELX(RelocGOTMixin, RelocTruncate32Mixin, GenericPCRelativeAddendReloc):
    check_sign_extend = True


class R_X86_64_REX_GOTPCRELX(RelocGOTMixin, RelocTruncate32Mixin, GenericPCRelativeAddendReloc):
    check_sign_extend = True


relocation_table_amd64 = {
    1: R_X86_64_64,
    2: R_X86_64_PC32,
    # 3: R_X86_64_GOT32,
    4: R_X86_64_PLT32,
    5: R_X86_64_COPY,
    6: R_X86_64_GLOB_DAT,
    7: R_X86_64_JUMP_SLOT,
    8: R_X86_64_RELATIVE,
    9: R_X86_64_GOTPCREL,
    10: R_X86_64_32,
    11: R_X86_64_32S,
    # 12: R_X86_64_16,
    # 13: R_X86_64_PC16,
    # 14: R_X86_64_8,
    # 15: R_X86_64_PC8,
    16: R_X86_64_DTPMOD64,
    17: R_X86_64_DTPOFF64,
    18: R_X86_64_TPOFF64,
    # 19: R_X86_64_TLSGD,
    # 20: R_X86_64_TLSLD,
    # 21: R_X86_64_DTPOFF32,
    # 22: R_X86_64_GOTTPOFF,
    # 23: R_X86_64_TPOFF32,
    # 24: R_X86_64_PC64,
    # 25: R_X86_64_GOTOFF64,
    # 26: R_X86_64_GOTPC32,
    # 32: R_X86_64_SIZE32,
    # 33: R_X86_64_SIZE64,
    # 34: R_X86_64_GOTPC32_TLSDESC,
    # 35: R_X86_64_TLSDESC_CALL,
    # 36: R_X86_64_TLSDESC,
    37: R_X86_64_IRELATIVE,
    # 38: R_X86_64_RELATIVE64,
    # 39, 40: Deprecated
    41: R_X86_64_GOTPCRELX,
    42: R_X86_64_REX_GOTPCRELX,
    # 43: R_X86_64_CODE_4_GOTPCRELX,
    # 44: R_X86_64_CODE_4_GOTTPOFF,
    # 45: R_X86_64_CODE_4_GOTPC32_TLSDESC,
    # 46: R_X86_64_CODE_5_GOTPCRELX,
    # 47: R_X86_64_CODE_5_GOTTPOFF,
    # 48: R_X86_64_CODE_5_GOTPC32_TLSDESC,
    # 49: R_X86_64_CODE_6_GOTPCRELX,
    # 50: R_X86_64_CODE_6_GOTTPOFF,
    # 51: R_X86_64_CODE_6_GOTPC32_TLSDESC,
}

__all__ = ("relocation_table_amd64",)



from __future__ import annotations

from typing import TYPE_CHECKING

from elftools.dwarf.die import DIE

from cle.address_translator import AT

from .variable_type import VariableType

if TYPE_CHECKING:
    from .elf import ELF
    from .subprogram import LexicalBlock


class Variable:
    """
    Variable for DWARF
    from a DW_TAG_variable or DW_TAG_formal_parameter

    :ivar str name:         The name of the variable
    :ivar relative_addr:    The relative addr (base addr depends on the type)
    :ivar lexical_block:    For a local variable, the lexical block where the variable is declared
    """

    def __init__(self, elf_object: ELF):
        self._elf_object = elf_object
        # all other optional params can be set afterwards
        self.relative_addr = None
        self.name = None
        self._type_offset = None
        self.decl_line = None
        self.decl_file = None
        self.lexical_block = None
        self.external = False
        self.declaration_only = False

    @staticmethod
    def from_die(die: DIE, expr_parser, elf_object: ELF, lexical_block: LexicalBlock | None = None):
        # first the address
        if "DW_AT_location" in die.attributes and die.attributes["DW_AT_location"].form == "DW_FORM_exprloc":
            parsed_exprs = expr_parser.parse_expr(die.attributes["DW_AT_location"].value)
            if len(parsed_exprs) == 1 and parsed_exprs[0].op_name == "DW_OP_addr":
                addr = parsed_exprs[0].args[0]
                var = MemoryVariable(elf_object, addr)
            elif len(parsed_exprs) == 1 and parsed_exprs[0].op_name == "DW_OP_fbreg":
                addr = parsed_exprs[0].args[0]
                var = StackVariable(elf_object, addr)
            elif len(parsed_exprs) == 1 and parsed_exprs[0].op_name.startswith("DW_OP_reg"):
                addr = parsed_exprs[0].op - 0x50  # 0x50 == DW_OP_reg0
                var = RegisterVariable(elf_object, addr)
            else:
                var = Variable(elf_object)
        else:
            var = Variable(elf_object)

        if "DW_AT_name" in die.attributes:
            var.name = die.attributes["DW_AT_name"].value.decode("utf-8")
        if "DW_AT_type" in die.attributes:
            var._type_offset = die.attributes["DW_AT_type"].value + die.cu.cu_offset
        if "DW_AT_decl_line" in die.attributes:
            var.decl_line = die.attributes["DW_AT_decl_line"].value
        if "DW_AT_external" in die.attributes:
            var.external = True
        if "DW_AT_declaration" in die.attributes:
            var.declaration_only = True

        var.lexical_block = lexical_block

        return var

    # overwritten for stack variables
    def rebased_addr_from_cfa(self, cfa: int):
        """
        The address of this variable in the global memory.

        :param cfa:     The canonical frame address as described by the DWARF standard.
        """
        return self.rebased_addr

    @property
    def rebased_addr(self):
        return None

    @property
    def addr(self):
        """
        Please use 'relative_addr' or 'rebased_addr' instead.
        """
        return self.relative_addr

    @property
    def type(self) -> VariableType:
        try:
            return self._elf_object.type_list[self._type_offset]
        except KeyError:
            return None

    @property
    def sort(self) -> str:
        # sort = 'stack' | 'register' | 'global'
        return "unknown"


class MemoryVariable(Variable):
    """
    This includes all variables that are not on the stack and not in a register.
    So all global variables, and also local static variables in C!
    """

    def __init__(self, elf_object: ELF, relative_addr):
        super().__init__(elf_object)
        self.relative_addr = relative_addr

    @property
    def rebased_addr(self):
        return AT.from_rva(self.relative_addr, self._elf_object).to_mva()

    @property
    def sort(self) -> str:
        return "global"


class StackVariable(Variable):
    """
    Stack Variable from DWARF.
    """

    def __init__(self, elf_object: ELF, relative_addr):
        super().__init__(elf_object)
        self.relative_addr = relative_addr

    def rebased_addr_from_cfa(self, cfa: int):
        return self.relative_addr + cfa

    @property
    def sort(self) -> str:
        return "stack"


class RegisterVariable(Variable):
    """
    Register Variable from DWARF.
    """

    def __init__(self, elf_object: ELF, register_addr):
        super().__init__(elf_object)
        # FIXME should this really go into relative addr?
        self.relative_addr = register_addr

    @property
    def sort(self) -> str:
        return "register"



"""Relocation types for MIPS 32-bit.

Reference: https://refspecs.linuxfoundation.org/elf/mipsabi.pdf page 4-19

The main document is old and does not contain all the relocation types. I
could not find a more recent document, so I had to rely on the source code of
GNU binutils for all relocations that are not in the main document. See
include/elf/mips.h in the binutils source code.
"""

from __future__ import annotations

from ctypes import c_int16

from archinfo import Endness

from .generic import (
    GenericAbsoluteAddendReloc,
    GenericAbsoluteReloc,
    GenericCopyReloc,
    GenericRelativeReloc,
    GenericTLSDoffsetReloc,
    GenericTLSModIdReloc,
    GenericTLSOffsetReloc,
)

# pylint: disable=missing-class-docstring


class R_MIPS_32(GenericAbsoluteAddendReloc):
    pass


class R_MIPS_REL32(GenericRelativeReloc):
    pass


class R_MIPS_26(GenericAbsoluteReloc):
    def relocate(self):
        if not self.resolved:
            return False

        original_value = self.owner.memory.unpack_word(self.dest_addr)
        original_value += self.value // 4

        self.owner.memory.pack_word(self.dest_addr, original_value)
        return True


class R_MIPS_JUMP_SLOT(GenericAbsoluteReloc):
    pass


class R_MIPS_GLOB_DAT(GenericAbsoluteReloc):
    pass


class R_MIPS_TLS_DTPMOD32(GenericTLSModIdReloc):
    pass


class R_MIPS_TLS_TPREL32(GenericTLSOffsetReloc):
    pass


class R_MIPS_TLS_DTPREL32(GenericTLSDoffsetReloc):
    pass


class R_MIPS_HI16(GenericAbsoluteReloc):

    def find_matching_lo16_relocation(self):
        current_hi16_index = self.owner.relocs.index(self)
        return next(
            reloc
            for reloc in self.owner.relocs[current_hi16_index:]
            if (self.symbol == reloc.symbol and type(reloc) is R_MIPS_LO16)
        )

    def relocate(self):
        if not self.resolved:
            return False

        # Relocating R_MIPS_HI16 requires to know the value placed at the following R_MIPS_LO16 relocation
        matching_lo16_reloc_dest_addr = self.find_matching_lo16_relocation().dest_addr

        dest_addr = self.dest_addr
        if self.arch.memory_endness == Endness.BE:
            dest_addr += 2
            matching_lo16_reloc_dest_addr += 2

        matching_lo16_reloc_target_bytes = c_int16(
            self.owner.memory.unpack_word(matching_lo16_reloc_dest_addr, size=2)
        ).value

        target_value = (self.value + matching_lo16_reloc_target_bytes) - (
            (self.value + matching_lo16_reloc_target_bytes) & 0xFFFF
        )
        target_value = (target_value >> 16) + self.owner.memory.unpack_word(dest_addr, size=2)

        self.owner.memory.pack_word(dest_addr, target_value, size=2)
        return True


class R_MIPS_LO16(GenericAbsoluteReloc):
    def relocate(self):
        if not self.resolved:
            return False

        dest_addr = self.dest_addr
        if self.arch.memory_endness == Endness.BE:
            dest_addr += 2

        target_value = (self.value + self.owner.memory.unpack_word(dest_addr, size=2)) & 0xFFFF

        self.owner.memory.pack_word(dest_addr, target_value, size=2)
        return True


class R_MIPS_64(GenericAbsoluteAddendReloc):
    pass


class R_MIPS_COPY(GenericCopyReloc):
    pass


class R_MIPS_TLS_DTPMOD64(GenericTLSModIdReloc):
    pass


class R_MIPS_TLS_DTPREL64(GenericTLSDoffsetReloc):
    pass


class R_MIPS_TLS_TPREL64(GenericTLSOffsetReloc):
    pass


relocation_table_mips = {
    # 1: R_MIPS_16,
    2: R_MIPS_32,
    3: R_MIPS_REL32,
    4: R_MIPS_26,
    5: R_MIPS_HI16,
    6: R_MIPS_LO16,
    # 7: R_MIPS_GPREL16,
    # 8: R_MIPS_LITERAL,
    # 9: R_MIPS_GOT16,
    # 10: R_MIPS_PC16,
    # 11: R_MIPS_CALL16,
    # 12: R_MIPS_GPREL32,
    # 13: R_MIPS_UNUSED1,
    # 14: R_MIPS_UNUSED2,
    # 15: R_MIPS_UNUSED3,
    # 16: R_MIPS_SHIFT5,
    # 17: R_MIPS_SHIFT6,
    18: R_MIPS_64,
    # 19: R_MIPS_GOT_DISP,
    # 20: R_MIPS_GOT_PAGE,
    # 21: R_MIPS_GOT_OFST,
    # 22: R_MIPS_GOT_HI16,
    # 23: R_MIPS_GOT_LO16,
    # 24: R_MIPS_SUB,
    # 25: R_MIPS_INSERT_A,
    # 26: R_MIPS_INSERT_B,
    # 27: R_MIPS_DELETE,
    # 28: R_MIPS_HIGHER,
    # 29: R_MIPS_HIGHEST,
    # 30: R_MIPS_CALL_HI16,
    # 31: R_MIPS_CALL_LO16,
    # 32: R_MIPS_SCN_DISP,
    # 33: R_MIPS_REL16,
    # 34: R_MIPS_ADD_IMMEDIATE,
    # 35: R_MIPS_PJUMP,
    # 36: R_MIPS_RELGOT,
    # 37: R_MIPS_JALR,
    38: R_MIPS_TLS_DTPMOD32,
    39: R_MIPS_TLS_DTPREL32,
    40: R_MIPS_TLS_DTPMOD64,
    41: R_MIPS_TLS_DTPREL64,
    # 42: R_MIPS_TLS_GD,
    # 43: R_MIPS_TLS_LDM,
    # 44: R_MIPS_TLS_DTPREL_HI16,
    # 45: R_MIPS_TLS_DTPREL_LO16,
    # 46: R_MIPS_TLS_GOTTPREL,
    47: R_MIPS_TLS_TPREL32,
    48: R_MIPS_TLS_TPREL64,
    # 49: R_MIPS_TLS_TPREL_HI16,
    # 50: R_MIPS_TLS_TPREL_LO16,
    51: R_MIPS_GLOB_DAT,
    # 60: R_MIPS_PC21_S2,
    # 61: R_MIPS_PC26_S2,
    # 62: R_MIPS_PC18_S3,
    # 63: R_MIPS_PC19_S2,
    # 64: R_MIPS_PCHI16,
    # 65: R_MIPS_PCLO16,
    126: R_MIPS_COPY,
    127: R_MIPS_JUMP_SLOT,
    # 248: R_MIPS_PC32,
    # 249: R_MIPS_EH,
    # 250: R_MIPS_GNU_REL16_S2,
    # 253: R_MIPS_GNU_VTINHERIT,
    # 254: R_MIPS_GNU_VTENTRY,
}

__all__ = ("relocation_table_mips",)



from __future__ import annotations

from .backend import Backend, register_backend
from .region import EmptySegment

__all__ = ("NamedRegion",)


class NamedRegion(Backend):
    """
    A NamedRegion represents a region of memory that has a name, a location, but no static content.

    This region also has permissions; with no memory, these obviously don't do anything on their own,
    but they help inform any other code that relies on CLE (e.g., angr)

    This can be used as a placeholder for memory that should exist in CLE's view, but for which it does not need data,
    like RAM, MMIO, etc
    """

    is_default = False  # This backend must be constructed manually (or by angr)
    has_memory = False  # This backend, by definition, has no memory backer

    def __init__(self, name, start, end, is_readable=True, is_writable=True, is_executable=False, **kwargs):
        """
        Create a NamedRegion.

        :param name: The name of the region
        :param start: The start address of the region
        :param end: The end address (exclusive) of the region
        :param is_readable: Whether the region is readable
        :param is_writable: Whether the region is writable
        :param is_executable: Whether the region is executable
        :param kwargs:
        """
        self.name = name
        super().__init__(name, None, **kwargs)
        self._min_addr = start
        self.linked_base = start
        self._max_addr = end
        self.has_memory = False
        s = EmptySegment(start, end - start, is_readable, is_writable, is_executable)
        self.segments.append(s)

    def __repr__(self):
        return f"<NamedRegion {self.name}, maps [{self.min_addr:#x}:{self.max_addr:#x}]>"

    @staticmethod
    def is_compatible(stream):
        return stream == 0  # I hate pylint

    @property
    def min_addr(self):
        return self._min_addr

    @property
    def max_addr(self):
        return self._max_addr

    def function_name(self, addr):  # pylint: disable=unused-argument,no-self-use
        """
        NamedRegions don't support function names.
        """
        return None

    def contains_addr(self, addr):
        return self.min_addr <= addr < self.max_addr

    @classmethod
    def check_compatibility(cls, spec, obj):  # pylint: disable=unused-argument
        return False


register_backend("named_region", NamedRegion)



from __future__ import annotations

import archinfo

from cle.backends.tls.tls_object import ThreadManager


class MinidumpThreadManager(ThreadManager):
    def __init__(self, loader, arch, **kwargs):  # pylint: disable=unused-argument
        self.loader = loader
        self.arch = arch
        self.threads = [
            MinidumpThread(loader, arch, loader.main_object.thread_registers(tid)) for tid in loader.main_object.threads
        ]
        self.modules = []  # ???

    def new_thread(self, insert=False):  # pylint: disable=no-self-use
        raise TypeError("Cannot create new threads from a minidump file... for now")

    def register_object(self, obj):
        pass


class MinidumpThread:
    def __init__(self, loader, arch: archinfo.Arch, registers):
        self.loader = loader
        self.arch = arch
        self._registers = registers
        if arch.name == "AMD64":
            self.teb = registers["gs_const"]
            self.thread_pointer = loader.main_object.memory.unpack_word(self.teb + 0x58)
        elif arch.name == "X86":
            self.teb = registers["fs"]
            self.thread_pointer = loader.main_object.memory.unpack_word(self.teb + 0x2C)

        self.user_thread_pointer = self.thread_pointer

    def get_tls_data_addr(self, tls_idx):
        return self.loader.memory.unpack_word(self.thread_pointer + tls_idx * self.arch.bytes)
